<?php
/**
 * i18nActiveRecord class file.
 *
 * @author K. O. Frimpong <me@kofrimpong.com>
 * @link http://www.kofrimpong.com/
 * @license New BSD License
 */

/**
 * i18nActiveRecord is the base class for classes representing internationalized relational data.
 *
 * It extends {@link CAtiveRecord} which implements the active record design pattern, a popular Object-Relational Mapping (ORM) technique.
 *
 * @property CDbCriteria $dbCriteria The query criteria that is associated with this model.
 * This criteria is mainly used by {@link scopes named scope} feature to accumulate
 * different criteria specifications.
 * @property CActiveRecordMetaData $metaData The meta for this AR class.
 * @property CDbConnection $dbConnection The database connection used by active record.
 * @property CDbTableSchema $tableSchema The metadata of the table that this AR belongs to.
 * @property CDbCommandBuilder $commandBuilder The command builder used by this AR.
 * @property array $attributes Attribute values indexed by attribute names.
 * @property boolean $isNewRecord Whether the record is new and should be inserted when calling {@link save}.
 * This property is automatically set in constructor and {@link populateRecord}.
 * Defaults to false, but it will be set to true if the instance is created using
 * the new operator.
 * @property mixed $primaryKey The primary key value. An array (column name=>column value) is returned if the primary key is composite.
 * If primary key is not defined, null will be returned.
 * @property mixed $oldPrimaryKey The old primary key value. An array (column name=>column value) is returned if the primary key is composite.
 * If primary key is not defined, null will be returned.
 * @property string $tableAlias The default table alias.
 *
 * @author  K. O. Frimpong <me@kofrimpong.com>
 * @version 1.0
 */
 
abstract class i18nActiveRecord extends CActiveRecord {

    /**
     *
     * @var array of translation objects indexed by the locale 
     */
    private $i18n = array();

    /**
     * Current locale
     * @var type 
     */
    private $_locale = null;
	
	private $_i18nAlias = 'tr';// the i18n table alias being used for query

    /**
     * Constructor.
     * @param string $scenario scenario name. See {@link CModel::scenario} for more details about this parameter.
     */
    public function __construct($scenario = 'insert') {

        if ($scenario === null) // internally used by populateRecord() and model()
            return;

        $this->setScenario($scenario);
        $this->setIsNewRecord(true);
        $this->_attributes = $this->getMetaData()->attributeDefaults;

        $this->init();

        $this->attachBehaviors($this->behaviors());
        $this->afterConstruct();
    }

    /**
     * Sets the locale to use
	 * If a local is not set, this object cannot return a property of its translation 
     * @param string $locale the CLocale id
     * @return i18nActiveRecord returns this active record
     */
    public function setLocale($locale) {
        $this->_locale = $locale;
        if (!isset($this->i18n[$this->_locale])) {
            $class = $this->getTranslationClass();
            if (!$this->getIsNewRecord()) {
                $record = self::model($this->getTranslationClass())->findByAttributes(array('i18n_id' => $this->getPrimaryKey(), 'locale' => $this->_locale));
                if ($record !== null) {
                    $record->setScenario($this->getScenario());
                    $this->i18n[$this->_locale] = $record;
                    return $this;
                }
                
            }
            $record = new $class();
            $record->locale = $this->_locale;
            $this->i18n[$this->_locale] = $record;
        }
        return $this;
    }
    
	/**
     * Returns the current locale for the active translation 
     * @return string CLocale id
     */
    public function getLocale(){
        return $this->_locale;
    }

    /**
     * Returns the translation for this locale
     * The translation record will be created if it does not exist
     * @param string $locale a CLocale id
     * @return  CActiveRecord the translation record
     */
    public function getTranslation($locale) {
        if (isset($this->i18n[$locale])) {
            return $this->i18n[$locale];
        }
        if(!$this->getIsNewRecord()){
            $record = self::model($this->getTranslationClass())->findByAttributes(array('i18n_id' => $this->getPrimaryKey(), 'locale' => $this->_locale));
            if($record !== null){
                $this->i18n[$locale] = $record;
            }
            return $record;
        }
        $class = $this->getTranslationClass();
        $record = new $class();
        $record->locale = $locale;
        $this->i18n[$locale] = $record;
        return $record;
    }

    /**
     * Deletes the translation ActiveRecord,
     * The translation ActiveRecord will be deleted from the database as well
     * @param string $locale a CLocale id
     * @return boolean whether deletion was successful 
     */
    public function deleteTranslation($locale) {
        if (isset($this->i18n[$locale])) {
            unset($this->i18n[$locale]);
        }
        return (boolean)self::model($this->getTranslationClass())->deleteAllByAttributes(array('i18n_id' => $this->getPrimaryKey(), 'locale' => $locale));
    }

    /**
     * Gets the name of the translation ActiveRecord class
     * Defaults to <code>get_class($this).'I18n'</code>
	 * @return string the class name of the translation record
     */
    public function getTranslationClass() {
        return get_class($this) . 'I18n';
    }

    /**
     * Adds a translation to this object
     * @param CAtiveRecord $i18nItem 
     */
    public function addTranslation($i18nItem) {
        if (isset($i18nItem->locale)) {
			$this->_locale = $i18nItem->locale;
            $this->i18n[$this->_locale] = $i18nItem;
        }
        else {
            throw new CException(Yii::t('yii', "'locale' attribute of translation object not set."));
        }
    }
	
    /**
     * Checks to see if a translation  exists for this locale
     * @param string $locale a CLocale id
     * @return boolean whether translation  exists
     */
	public function translationExists($locale){
		 if (isset($this->i18n[$locale])) {
            return true;
        }
		if(!$this->getIsNewRecord()){
			return self::model($this->getTranslationClass())->exists('i18n_id=:i18n_id AND locale=:locale',array('i18n_id' => $this->getPrimaryKey(), 'locale' => $locale));
		}
		return false;
	}
    
    /**
     * Calls an annonymous function on each translation 
     *The signature of the closure should be <code>function($translationObject){}</code>.
     * @param Closure annonymous function
     */
    public function eachTranslation(Closure $closure){
        foreach ($this->i18n as $i18nItem) {
            $closure($i18nItem);
        }
    }

  /**
     * A named scope to fetch translation together with this record
     * @param string $local a CLocale id
     * @param string $join the join type to user
     * use LEFT OUTER JOIN to get all records including those without translation 
     * use INNER JOIN to return only records with translation 
     * Default to LEFT OUTER JOIN
	 * @return i18nActiveRecord this record
     */
    public function withI18n($locale, $join = 'LEFT OUTER JOIN') {
        $this->_locale = $locale;
		//the following cumbersome process is to trick CActiveFinder getColumnSelect() not to strip the select statement (tr.*) for i18n table when doing relational query
		$i18nTable = self::model($this->getTranslationClass())->getTableSchema();
		$columns = array();
		$schema = $this->dbConnection->schema;
		$alias = $schema->quoteColumnName($this->_i18nAlias);
		foreach($i18nTable->getColumnNames() as $name){
			$n = $schema->quoteColumnName($name);
			$columns[]="{$alias}.{$n} AS {$name}";
		}
		$columns = implode(', ',$columns);
        $this->getDbCriteria()->mergeWith(array(
            'select' => "t.*,{$columns}",
            'join' => "{$join} {$i18nTable->name} {$alias} ON (t.id = {$alias}.i18n_id AND {$alias}.locale = :locale)",
            'params' => array(':locale' => $this->_locale)
        ));
        return $this;
    }
    
    /**
     * A named scope for searching records  using a condition on the translation
     *The alias tr is given for the translation table
     * @param string $local a CLocale id
     * @param string $condition the condition of the translation
     * @params array $params parameters to be bound to the condition
	 * @return i18nActiveRecord this record
     */
    public function usingI18n($locale, $condition, $params=array()) {
        $i18nTable = self::model($this->getTranslationClass())->tableName();
        $params[':_locale'] = $locale;
        $this->getDbCriteria()->mergeWith(array(
            'join' => "INNER JOIN {$i18nTable} AS {$this->_i18nAlias} ON {$this->_i18nAlias}.locale = :_locale",
            'condition'=>$condition,
            'params'=>$params
        ));
        return $this;
    }


    /**
     * PHP getter magic method.
     * This method is overridden so that AR attributes can be accessed like properties.
     * @param string $name property name
     * @return mixed property value
     * @see getAttribute
     */
    public function __get($name) {
        if (isset($this->_attributes[$name]))
            return $this->_attributes[$name];
        else if (isset($this->i18n[$this->_locale]) && isset($this->i18n[$this->_locale]->_attributes[$name]))
            return $this->i18n[$this->_locale]->_attributes[$name];
        else if (isset($this->getMetaData()->columns[$name]))//not populated
            return null;
        else if (isset($this->i18n[$this->_locale]) && isset($this->i18n[$this->_locale]->getMetaData()->columns[$name]))
            return null;
        else if (isset($this->_related[$name]))
            return $this->_related[$name];
        else if (isset($this->getMetaData()->relations[$name]))
            return $this->getRelated($name);
        else
            return parent::__get($name);
    }

    /**
     * PHP setter magic method.
     * This method is overridden so that AR attributes can be accessed like properties.
     * @param string $name property name
     * @param mixed $value property value
     */
    public function __set($name, $value) {
        if ($this->setAttribute($name, $value) === false) {
            if (isset($this->getMetaData()->relations[$name]))
                $this->_related[$name] = $value;
            else
                parent::__set($name, $value);
        }
    }

    /**
     * Checks if a property value is null.
     * This method overrides the parent implementation by checking
     * if the named attribute is null or not.
     * @param string $name the property name or the event name
     * @return boolean whether the property value is null
     */
    public function __isset($name) {
        if (isset($this->_attributes[$name]))
            return true;
        else if (isset($this->i18n[$this->_locale]) && isset($this->i18n[$this->_locale]->_attributes[$name]))
            return true;
        else if (isset($this->getMetaData()->columns[$name]))
            return false;
        else if (isset($this->_related[$name]))
            return true;
        else if (isset($this->getMetaData()->relations[$name]))
            return $this->getRelated($name) !== null;
        else
            return parent::__isset($name);
    }

    /**
     * Sets a component property to be null.
     * This method overrides the parent implementation by clearing
     * the specified attribute value.
     * @param string $name the property name or the event name
     */
    public function __unset($name) {
        if (isset($this->getMetaData()->columns[$name]))
            unset($this->_attributes[$name]);
        if (isset($this->i18n[$this->_locale]) && isset($this->i18n[$this->_locale]->getMetaData()->columns[$name]))
            unset($this->i18n[$this->_locale]->_attributes[$name]);
        else if (isset($this->getMetaData()->relations[$name]))
            unset($this->_related[$name]);
        else
            parent::__unset($name);
    }

    /**
     * Checks whether this AR has the named attribute
     * @param string $name attribute name
     * @return boolean whether this AR has the named attribute (table column).
     */
    public function hasAttribute($name) {
        return parent::hasAttribute($name) || self::model($this->getTranslationClass())->hasAttribute($name);
    }

    /**
     * Returns the named attribute value.
     * If this is a new record and the attribute is not set before,
     * the default column value will be returned.
     * If this record is the result of a query and the attribute is not loaded,
     * null will be returned.
     * You may also use $this->AttributeName to obtain the attribute value.
     * @param string $name the attribute name
     * @return mixed the attribute value. Null if the attribute is not set or does not exist.
     * @see hasAttribute
     */
    public function getAttribute($name) {
        if (($value = parent::getAttribute($name) ) === null) {
            return self::model($this->getTranslationClass())->getAttribute($name);
        }
        return $value;
    }

    /**
     * Sets the named attribute value.
     * You may also use $this->AttributeName to set the attribute value.
     * @param string $name the attribute name
     * @param mixed $value the attribute value.
     * @return boolean whether the attribute exists and the assignment is conducted successfully
     * @see hasAttribute
     */
    public function setAttribute($name, $value) {
        if (!parent::setAttribute($name, $value)) {
            return isset($this->i18n[$this->_locale]) && $this->i18n[$this->_locale]->setAttribute($name, $value);
        }
        return true;
    }

    /**
     * Returns all column attribute values.
     * Note, related objects are not returned.
     * @param mixed $names names of attributes whose value needs to be returned.
     * If this is true (default), then all attribute values will be returned, including
     * those that are not loaded from DB (null will be returned for those attributes).
     * If this is null, all attributes except those that are not loaded from DB will be returned.
     * @return array attribute values indexed by attribute names.
     */
    public function getAttributes($names = true) {
        $i18nAttributes = self::model($this->getTranslationClass())->getAttributes($names);
        return array_merge($i18nAttributes, parent::getAttributes($names = true));
    }

    /**
     * Returns the list of all attribute names of the model.
     * This would return all column names of the table associated with this AR class.
     * @return array list of attribute names.
     */
    public function attributeNames() {
        return array_merge(array_keys(self::model($this->getTranslationClass())->getMetaData()->columns), parent::attributeNames());
    }

    /**
     * Returns the text label for the specified attribute.
     * @param string $attribute the attribute name
     * @return string the attribute label
     * @see generateAttributeLabel
     * @see attributeLabels
     */
    public function getAttributeLabel($attribute) {
        $labels = array_merge(CActiveRecord::model($this->getTranslationClass())->attributeLabels(), $this->attributeLabels());
        if (isset($labels[$attribute]))
            return $labels[$attribute];
        else
            return $this->generateAttributeLabel($attribute);
    }

    /**
     * Saves the current record together with translation if any.
     *
     * The record is inserted as a row into the database table if its {@link isNewRecord}
     * property is true (usually the case when the record is created using the 'new'
     * operator). Otherwise, it will be used to update the corresponding row in the table
     * (usually the case if the record is obtained using one of those 'find' methods.)
     *
     * Validation will be performed before saving the record. If the validation fails,
     * the record will not be saved. You can call {@link getErrors()} to retrieve the
     * validation errors.
     *
     * If the record is saved via insertion, its {@link isNewRecord} property will be
     * set false, and its {@link scenario} property will be set to be 'update'.
     * And if its primary key is auto-incremental and is not set before insertion,
     * the primary key will be populated with the automatically generated key value.
     *
     * @param boolean $runValidation whether to perform validation before saving the record.
     * If the validation fails, the record will not be saved to database.
     * @param array $attributes list of attributes that need to be saved. Defaults to null,
     * meaning all attributes that are loaded from DB will be saved.
     * @return boolean whether the saving succeeds
     */
    public function save($runValidation = true, $attributes = null) {
        if (!$runValidation || $this->validate($attributes)) {
            $pdo = $this->dbConnection->getPdoInstance();
            if (!$pdo->inTransaction()) {
                $pdo->beginTransaction();
                try {
                    $this->saveInternal($attributes);
                    return $pdo->commit();
                }
                catch (Exception $e) {
                    $pdo->rollback();
                    throw new CDbException(Yii::t('yii', $e->getMessage()));
                }
            }
            else {
                return $this->saveInternal($attributes);
            }
        }
        else
            return false;
    }

    private function saveInternal($attributes) {
        if ($this->getIsNewRecord()) {
            $result = $this->insert($attributes);
            $_id = $this->getPrimaryKey();
            foreach ($this->i18n as $i18nItem) {
                $i18nItem->i18n_id = $_id;
                $result = $i18nItem->insert($attributes) && $result;
            }
        }
        else {
            $result = $this->update($attributes);
            $_id = $this->getPrimaryKey();
            foreach ($this->i18n as $i18nItem) {
                if ($i18nItem->getIsNewRecord()) {
                    $i18nItem->i18n_id = $_id;
                    $result = $i18nItem->insert($attributes) && $result;
                }
                else {
                    $result = $i18nItem->update($attributes) && $result;
                }
            }
        }
        return $result;
    }

    /**
     * Saves one or several counter columns for the current AR object.
     * Note that this method differs from {@link updateCounters} in that it only
     * saves the current AR object.
     * An example usage is as follows:
     * <pre>
     * $postRecord=Post::model()->findByPk($postID);
     * $postRecord->saveCounters(array('view_count'=>1));
     * </pre>
     * Use negative values if you want to decrease the counters.
     * @param array $counters the counters to be updated (column name=>increment value)
     * @return boolean whether the saving is successful
     * @see updateCounters
     * @since 1.1.8
     */
    public function saveCounters($counters) {
        return parent::saveCounters($counters) && (isset($this->i18n[$this->_locale]) && $this->i18n[$this->_locale]->saveCounters($counters));
    }

    /**
     * Saves a selected list of attributes.
     * Unlike {@link save}, this method only saves the specified attributes
     * of an existing row dataset and does NOT call either {@link beforeSave} or {@link afterSave}.
     * Also note that this method does neither attribute filtering nor validation.
     * So do not use this method with untrusted data (such as user posted data).
     * You may consider the following alternative if you want to do so:
     * <pre>
     * $postRecord=Post::model()->findByPk($postID);
     * $postRecord->attributes=$_POST['post'];
     * $postRecord->save();
     * </pre>
     * @param array $attributes attributes to be updated. Each element represents an attribute name
     * or an attribute value indexed by its name. If the latter, the record's
     * attribute will be changed accordingly before saving.
     * @return boolean whether the update is successful
     * @throws CException if the record is new or any database error
     */
    public function saveAttributes($attributes) {
        return parent::saveAttribute($attributes) && (isset($this->i18n[$this->_locale]) && $this->i18n[$this->_locale]->saveAttributes($attributes));
    }

    /**
     * Deletes the row corresponding to this active record.
     * @return boolean whether the deletion is successful.
     * @throws CException if the record is new
     */
    public function delete() {
        if (!$this->getIsNewRecord()) {
            Yii::trace(get_class($this) . '.delete()', 'system.db.ar.CActiveRecord');
            if ($this->beforeDelete()) {
                $pdo = $this->dbConnection->getPdoInstance();
                if (!$pdo->inTransaction()) {
                    $pdo->beginTransaction();
                    try {
                        $this->deleteInternal();
                        $result = $pdo->commit();
                    }
                    catch (Exception $e) {
                        $pdo->rollback();
                        throw new CDbException(Yii::t('yii', $e->getMessage()));
                    }
                }
                else {
                    $result = $this->deleteInternal();
                }
                $this->afterDelete();
                unset($this->i18n);
                return $result;
            }
            else
                return false;
        }
        else
            throw new CDbException(Yii::t('yii', 'The active record cannot be deleted because it is new.'));
    }

    private function deleteInternal() {
        return $this->deleteByPk($p) && self::model($this->getTranslationClass())->deleteAll("i18n_id=:id",array(':id'=>$this->getPrimaryKey()));
    }

    /**
     * Creates an active record with the given attributes.
     * This method is internally used by the find methods.
     * @param array $attributes attribute values (column name=>column value)
     * @param boolean $callAfterFind whether to call {@link afterFind} after the record is populated.
     * @return CActiveRecord the newly created active record. The class of the object is the same as the model class.
     * Null is returned if the input data is false.
     */
    public function populateRecord($attributes, $callAfterFind = true) {
        if ($attributes !== false) {
            if($this->_locale !== null){
                return $this->populateWithi18n($attributes, $callAfterFind);
            }
            return $this->populateRecordOnly($attributes, $callAfterFind);
        }
        else
            return null;
    }
    
    /**
     *
     * @param type $attributes
     * @param type $callAfterFind
     * @return CActiveRecord 
     */
    private function populateRecordOnly($attributes, $callAfterFind){
        $record = $this->instantiate($attributes);
            $record->setScenario('update');
            $record->init();
            $md = $record->getMetaData();
            foreach ($attributes as $name => $value) {
                if (property_exists($record, $name))
                    $record->$name = $value;
                else if (isset($md->columns[$name]))
                    $record->_attributes[$name] = $value;
            }
            $record->_pk = $record->getPrimaryKey();
            $record->attachBehaviors($record->behaviors());
            if ($callAfterFind)
                $record->afterFind();
            return $record;
    }
    
     /**
     *
     * @param type $attributes
     * @param type $callAfterFind
     * @return CActiveRecord 
     */
    private function populateWithi18n($attributes, $callAfterFind){ 
        $record = $this->instantiate($attributes);
            $record->setScenario('update');
            $record->_locale = $this->_locale;
            $record->init();
            $md = $record->getMetaData();
            
            $class = $this->getTranslationClass();
            $i18nItem = new $class(null);
            $i18nItem->setScenario('update');
            $mdI18n = $i18nItem->getMetaData();
            
            foreach ($attributes as $name => $value) {
                if (property_exists($record, $name))
                    $record->$name = $value;
                else if (isset($md->columns[$name]))
                    $record->_attributes[$name] = $value;
                else if (property_exists($i18nItem, $name))
                    $i18nItem->$name = $value;
                else if (isset($mdI18n->columns[$name]))
                    $i18nItem->_attributes[$name] = $value;
            }
            $record->_pk = $record->getPrimaryKey();
            $i18nItem->i18n_id = $record->_pk;
            $record->i18n[$record->_locale] = $i18nItem;
            $record->attachBehaviors($record->behaviors());
            if ($callAfterFind)
                $record->afterFind();
            return $record;
    }


    /**
     * Removes errors for all attributes or a single attribute.
     * @param string $attribute attribute name. Use null to remove errors for all attribute.
     */
    public function clearErrors($attribute = null) {
        parent::clearErrors($attribute);
        foreach($this->i18n as $i18nItem){
            $i18nItem->clearErrors($attribute);
        }
    }

    /**
     * Returns a value indicating whether there is any validation error.
     * @param string $attribute attribute name. Use null to check all attributes.
     * @return boolean whether there is any error.
     */
    public function hasErrors($attribute = null) {
        $hasError = parent::hasErrors($attribute);
        foreach($this->i18n as $i18nItem){
            $hasError = $hasError || $i18nItem->hasErrors($attribute);
        }
        return $hasError;
    }

    /**
     * Returns the errors for all attribute or a single attribute.
     * @param string $attribute attribute name. Use null to retrieve errors for all attributes.
     * @return array errors for all attributes or the specified attribute. Empty array is returned if no error.
     */
    public function getErrors($attribute = null) {
        return array_merge( isset($this->i18n[$this->_locale])?$this->i18n[$this->_locale]->getErrors($attribute):array(), parent::getErrors($attribute));
    }

    /**
     * Returns the first error of the specified attribute.
     * @param string $attribute attribute name.
     * @return string the error message. Null is returned if no error.
     */
    public function getError($attribute) {
        if (($value = parent::getError($attribute)) === null) {
            if(isset($this->i18n[$this->_locale])){
                return $this->i18n[$this->_locale]->getError($attribute);
            }
        }
        return $value;
    }
    
    	/**
	 * Performs the validation.
	 *
	 * This method executes the validation rules as declared in {@link rules}.
	 * Only the rules applicable to the current {@link scenario} will be executed.
	 * A rule is considered applicable to a scenario if its 'on' option is not set
	 * or contains the scenario.
	 *
	 * Errors found during the validation can be retrieved via {@link getErrors}.
	 *
	 * @param array $attributes list of attributes that should be validated. Defaults to null,
	 * meaning any attribute listed in the applicable validation rules should be
	 * validated. If this parameter is given as a list of attributes, only
	 * the listed attributes will be validated.
	 * @param boolean $clearErrors whether to call {@link clearErrors} before performing validation
	 * @return boolean whether the validation is successful without any error.
	 * @see beforeValidate
	 * @see afterValidate
	 */
	public function validate($attributes=null, $clearErrors=true)
	{
		if($clearErrors)
			$this->clearErrors();
		if($this->beforeValidate())
		{           
			foreach(parent::getValidators() as $validator){
				$validator->validate($this,$attributes);
            }
            foreach(self::model($this->getTranslationClass())->getValidators() as $validator){
                foreach($this->i18n as $i18nItem){
                    $validator->validate($i18nItem,$attributes);
                }
            }
			$this->afterValidate();
			return !$this->hasErrors();
		}
		else
			return false;
	}

    /**
     * Creates validator objects based on the specification in {@link rules}.
     * This method is mainly used internally.
     * @return CList validators built based on {@link rules()}.
     */
    public function createValidators() {
        $v = self::model($this->getTranslationClass())->createValidators();
        $v->mergeWith(parent::createValidators());
        return $v;
    }

}